@@ -72,17 +72,32 @@ STATIC int do_determin_dev_size(struct Drbd_Conf* mdev);
 int drbd_determin_dev_size(struct Drbd_Conf* mdev)
 {
 	sector_t pmdss; // previous meta data start sector
+	sector_t la_size;
+	int md_moved, la_size_changed;
 	int rv;
 
 	wait_event(mdev->al_wait, lc_try_lock(mdev->act_log));
 	pmdss = drbd_md_ss(mdev);
+	la_size = mdev->la_size;
+
 	rv = do_determin_dev_size(mdev);
-	if ( pmdss != drbd_md_ss(mdev) && mdev->md_index == -1 ) {
+	if (rv < 0) goto out;
+
+	la_size_changed = (la_size != mdev->la_size);
+	md_moved = pmdss != drbd_md_ss(mdev) /* && mdev->md_index == -1 */;
+
+	if ( md_moved ) {
 		WARN("Moving meta-data.\n");
+		D_ASSERT(mdev->md_index == -1);
+	}
+
+	if ( la_size_changed || md_moved ) {
 		drbd_al_shrink(mdev); // All extents inactive.
 		drbd_bm_write(mdev);  // write bitmap
-		drbd_md_write(mdev);  // Write mdev->la_size to disk.
+		// Write mdev->la_size to [possibly new position on] disk.
+		drbd_md_write(mdev);
 	}
+  out:
 	lc_unlock(mdev->act_log);
 
 	return rv;
@@ -103,7 +118,9 @@ char* ppsize(char* buf, size_t size)
 }
 
 
-/* Returns 1 if there is a disk-less node, 0 if both nodes have a disk. */
+/* Returns 1 if there is a disk-less node, 0 if both nodes have a disk.
+ * -ENOMEM if we could not allocate the bitmap
+ */
 /*
  * THINK do we want the size to be KB or sectors ?
  * note, *_capacity operates in 512 byte sectors!!
@@ -166,16 +183,24 @@ STATIC int do_determin_dev_size(struct Drbd_Conf* mdev)
 		int err;
 		err = drbd_bm_resize(mdev,size<<1); // wants sectors
 		if (unlikely(err)) {
-			ERR("BM resizing failed. "
-			    "Leaving size unchanged at size = %lu KB\n", 
-			    (unsigned long)size);
-		} else {
-			// racy, see comments above.
-			drbd_set_my_capacity(mdev,size<<1);
-			mdev->la_size = size;
-			INFO("size = %s (%lu KB)\n",ppsize(ppb,size),
-			     (unsigned long)size);
+			/* currently there is only one error: ENOMEM! */
+			size = drbd_bm_capacity(mdev)>>1;
+			if (size == 0) {
+				ERR("OUT OF MEMORY! Could not allocate bitmap! Set device size => 0\n");
+			} else {
+				/* FIXME this is problematic,
+				 * if we in fact are smaller now! */
+				ERR("BM resizing failed. "
+				    "Leaving size unchanged at size = %lu KB\n", 
+				    (unsigned long)size);
+			}
+			rv = err;
 		}
+		// racy, see comments above.
+		drbd_set_my_capacity(mdev,size<<1);
+		mdev->la_size = size;
+		INFO("size = %s (%lu KB)\n",ppsize(ppb,size),
+		     (unsigned long)size);
 	}
 
 	return rv;
@@ -229,9 +254,12 @@ STATIC int drbd_check_al_size(drbd_dev *mdev)
 	} else {
 		if (t) lc_free(t);
 	}
+	drbd_md_write(mdev);
 	return 0;
 }
 
+STATIC int drbd_detach_ioctl(drbd_dev *mdev);
+
 STATIC
 int drbd_ioctl_set_disk(struct Drbd_Conf *mdev,
 			struct ioctl_disk_config * arg)
@@ -285,6 +313,9 @@ int drbd_ioctl_set_disk(struct Drbd_Conf *mdev,
 		__module_get(THIS_MODULE);
 		mput = 1;
 	} else {
+		/* We currently cannot handle reattach while connected */
+		return -EBUSY;
+
 		/* FIXME allow reattach while connected,
 		 * and allow it in Primary/Diskless state...
 		 * currently there are strange races leading to a distributed
@@ -455,10 +486,22 @@ ONLY_IN_26({
 /* FIXME if (md_gc_valid < 0) META DATA IO NOT POSSIBLE! */
 
 	drbd_bm_lock(mdev); // racy...
-	drbd_determin_dev_size(mdev);
-	/* FIXME
-	 * what if we now have la_size == 0 ?? eh?
-	 */
+	if (drbd_determin_dev_size(mdev) < 0) {
+		/* could not allocate bitmap.
+		 * try to undo ... */
+		D_ASSERT(mdev->cstate == Unconfigured);
+		D_ASSERT(mput == 1);
+
+		drbd_bm_unlock(mdev);
+
+		/* from drbd_detach_ioctl */
+		drbd_free_ll_dev(mdev);
+
+		set_cstate(mdev,Unconfigured);
+		drbd_mdev_cleanup(mdev);
+		module_put(THIS_MODULE);
+		return -ENOMEM;
+	}
 
 	if (md_gc_valid <= 0) {
 		INFO("Assuming that all blocks are out of sync (aka FullSync)\n");
@@ -473,7 +516,6 @@ ONLY_IN_26({
 
 	i = drbd_check_al_size(mdev);
 	if (i) {
-// FATAL!
 		/* FIXME see the comment above.
 		 * if this fails I need to undo all changes,
 		 * go back into Unconfigured,
@@ -481,7 +523,6 @@ ONLY_IN_26({
 		 */
 		// return i;
 		drbd_panic("Cannot allocate act_log\n");
-		drbd_suicide();
 	}
 
 	if (md_gc_valid > 0) {
@@ -710,6 +751,8 @@ int drbd_set_state(drbd_dev *mdev,Drbd_State newstate)
 		 * become Secondary. */
 		if (bd_claim(mdev->this_bdev,drbd_sec_holder))
 			return -EBUSY;
+		if (disable_bd_claim)
+			bd_release(mdev->this_bdev);
 	}
 #endif
 
@@ -851,13 +894,27 @@ static int drbd_get_wait_time(long *tp, struct Drbd_Conf *mdev,
 	if(copy_from_user(&p,arg,sizeof(p))) {
 		return -EFAULT;
 	}
-
-	if( drbd_md_test_flag(mdev,MDF_ConnectedInd) ) {
-		time=p.wfc_timeout;
-		//ERR("using wfc_timeout.\n");
-	} else {
+	/* If I am currently not Primary,
+	 * but meta data primary indicator is set,
+	 * I just now recover from a hard crash,
+	 * and have been Primary before that crash.
+	 *
+	 * Now, if I had no connection before that crash
+	 * (have been degraded Primary), chances are that
+	 * I won't find my peer now either.
+	 *
+	 * In that case, and _only_ in that case,
+	 * we use the degr-wfc-timeout instead of the default,
+	 * so we can automatically recover from a crash of a
+	 * degraded but active "cluster" after a certain timeout.
+	 */
+	if ( mdev->state != Primary &&
+	     drbd_md_test_flag(mdev,MDF_PrimaryInd) &&
+	    !drbd_md_test_flag(mdev,MDF_ConnectedInd) ) {
 		time=p.degr_wfc_timeout;
-		//ERR("using degr_wfc_timeout.\n");
+		if (time) WARN("using degr_wfc_timeout=%ld seconds\n", time);
+	} else {
+		time=p.wfc_timeout;
 	}
 
 	time=time*HZ;
@@ -964,6 +1021,17 @@ STATIC int drbd_detach_ioctl(drbd_dev *mdev)
 	return 0;
 }
 
+#ifdef CONFIG_COMPAT
+long drbd_compat_ioctl(struct file *f, unsigned cmd, unsigned long arg)
+{
+	int ret;
+	// lock_kernel(); Not needed, since we have mdev->device_mutex
+	ret = drbd_ioctl(f->f_dentry->d_inode, f, cmd, arg);
+	// unlock_kernel();
+	return ret;
+}
+#endif
+
 int drbd_ioctl(struct inode *inode, struct file *file,
 			   unsigned int cmd, unsigned long arg)
 {
@@ -1105,10 +1173,26 @@ ONLY_IN_26(
 		} else {
 			clear_bit(ON_PRI_INC_HUMAN,&mdev->flags);
 			clear_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
+			if (arg == 0) break;
 
-			if (arg & Human ) 
+			// XXX reduce race: don't set it,
+			// if we have a connection.
+			// this does not avoid the race completely, though.
+			if (mdev->cstate > WFConnection) {
+				WARN("race avoidance: did not set "
+				     "the state flags (%s), cstate=%s\n",
+				        arg == (Human|TimeoutExpired)
+				     ?  "Human|TimeoutExpired"
+				     : arg == Human
+				     ? "Human"
+				     : "TimeoutExpired",
+				     cstate_to_name(mdev->cstate));
+				break;
+			}
+
+			if (arg & Human)
 				set_bit(ON_PRI_INC_HUMAN,&mdev->flags);
-			if (arg & TimeoutExpired )
+			if (arg & TimeoutExpired)
 				set_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
 		}
 		break;
@@ -1245,15 +1329,19 @@ ONLY_IN_26(
 		/* disallow "invalidation" of local replica
 		 * when currently in primary state (would be a Bad Idea),
 		 * or during a running sync (won't make any sense) */
-		if( mdev->state == Primary ||
-		    mdev->cstate < StandAlone ||
-		    mdev->cstate > Connected ||
+		if( (mdev->state == Primary ||
+		      (mdev->cstate != Connected &&
+		       mdev->cstate != StandAlone)) ||
 		    test_bit(DISKLESS,&mdev->flags) ||
 		    test_bit(PARTNER_DISKLESS,&mdev->flags) ) {
 			err = -EINPROGRESS;
 			break;
 		}
 
+		drbd_md_set_flag(mdev,MDF_FullSync);
+		drbd_md_clear_flag(mdev,MDF_Consistent);
+		drbd_md_write(mdev);
+
 		if (mdev->cstate == Connected) {
 			/* avoid races with set_in_sync
 			 * for successfull mirrored writes
@@ -1265,10 +1353,6 @@ ONLY_IN_26(
 
 		drbd_bm_lock(mdev); // racy...
 
-		drbd_md_set_flag(mdev,MDF_FullSync);
-		drbd_md_clear_flag(mdev,MDF_Consistent);
-		drbd_md_write(mdev);
-
 		drbd_bm_set_all(mdev);
 		drbd_bm_write(mdev);
 
@@ -1331,4 +1415,3 @@ ONLY_IN_26(
  out_unlocked:
 	return err;
 }
-
