@@ -407,7 +407,6 @@ STATIC int _drbd_process_ee(drbd_dev *mdev, int be_sleepy)
 
 	if( test_and_set_bit(PROCESS_EE_RUNNING,&mdev->flags) ) {
 		if(!be_sleepy) {
-			clear_bit(PROCESS_EE_RUNNING,&mdev->flags);
 			return 3;
 		}
 		spin_unlock_irq(&mdev->ee_lock);
@@ -615,15 +614,37 @@ STATIC struct socket *drbd_try_connect(drbd_dev *mdev)
 {
 	int err;
 	struct socket *sock;
+	struct sockaddr_in src_in;
 
 	err = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
 	if (err) {
 		ERR("sock_creat(..)=%d\n", err);
+		return NULL;
 	}
 
 	sock->sk->SK_(rcvtimeo) =
 	sock->sk->SK_(sndtimeo) =  mdev->conf.try_connect_int*HZ;
 
+       /* explicitly bind to the configured IP as source IP 
+	   for the outgoing connections.
+	   This is needed for multihomed hosts and to be 
+	   able to use lo: interfaces for drbd.
+          Make sure to use 0 as portnumber, so linux selects
+	   a free one dynamically.
+	*/
+	memcpy (&src_in, &(mdev->conf.my_addr), sizeof(struct sockaddr_in));
+	src_in.sin_port = 0; 
+
+	err = sock->ops->bind(sock,
+			      (struct sockaddr * ) &src_in,
+			      sizeof (struct sockaddr_in));
+	if (err) {
+		ERR("Unable to bind source sock (%d)\n", err);
+		sock_release(sock);
+		sock = NULL;
+		return sock;
+	}
+
 	err = sock->ops->connect(sock,
 				 (struct sockaddr *) mdev->conf.other_addr,
 				 mdev->conf.other_addr_len, 0);
@@ -643,7 +664,7 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 	err = sock_create(AF_INET, SOCK_STREAM, 0, &sock2);
 	if (err) {
 		ERR("sock_creat(..)=%d\n", err);
-		// FIXME return NULL ?
+		return NULL;
 	}
 
 	sock2->sk->SK_(reuse)    = 1; /* SO_REUSEADDR */
@@ -654,7 +675,7 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 			      (struct sockaddr *) mdev->conf.my_addr,
 			      mdev->conf.my_addr_len);
 	if (err) {
-		ERR("Unable to bind (%d)\n", err);
+		ERR("Unable to bind sock2 (%d)\n", err);
 		sock_release(sock2);
 		set_cstate(mdev,Unconnected);
 		return 0;
@@ -668,9 +689,17 @@ STATIC struct socket *drbd_wait_for_connect(drbd_dev *mdev)
 
 STATIC int drbd_do_handshake(drbd_dev *mdev);
 
+/*
+ * return values:
+ *   1 yess, we have a valid connection
+ *   0 oops, did not work out, please try again
+ *  -1 peer talks different language,
+ *     no point in trying again, please go standalone.
+ */
 int drbd_connect(drbd_dev *mdev)
 {
 	struct socket *sock,*msock;
+	int h;
 
 	D_ASSERT(mdev->cstate!=Unconfigured);
 	D_ASSERT(!mdev->data.socket);
@@ -699,12 +728,12 @@ int drbd_connect(drbd_dev *mdev)
 				sock_release(sock);
 			}
 		}
-		if(mdev->cstate==Unconnected) return 0;
+		if(mdev->cstate==Unconnected) return -1;
 		if(signal_pending(current)) {
 			drbd_flush_signals(current);
 			smp_rmb();
 			if (get_t_state(&mdev->receiver) == Exiting)
-				return 0;
+				return -1;
 		}
 	}
 
@@ -746,9 +775,8 @@ int drbd_connect(drbd_dev *mdev)
 	set_cstate(mdev,WFReportParams);
 	D_ASSERT(mdev->asender.task == NULL);
 
-	if (!drbd_do_handshake(mdev)) {
-		return 0;
-	}
+	h = drbd_do_handshake(mdev);
+	if (h <= 0) return h;
 
 	clear_bit(ON_PRI_INC_HUMAN,&mdev->flags);
 	clear_bit(ON_PRI_INC_TIMEOUTEX,&mdev->flags);
@@ -906,7 +934,7 @@ STATIC int e_end_resync_block(drbd_dev *mdev, struct drbd_work *w, int unused)
 			 */
 		}
 		ok = drbd_send_ack(mdev,WriteAck,e);
-		__set_bit(SYNC_STARTED,&mdev->flags);
+		set_bit(SYNC_STARTED,&mdev->flags);
 	} else {
 		ok = drbd_send_ack(mdev,NegAck,e);
 		ok&= drbd_io_error(mdev);
@@ -1348,8 +1376,8 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 {
 	Drbd_Parameter_Packet *p = (Drbd_Parameter_Packet*)h;
 	int consider_sync;
-	int oo_state;
-	sector_t p_size;
+	int oo_state,i;
+	sector_t p_size, p_usize;
 
 	if (h->length != (sizeof(*p)-sizeof(*h))) {
 		ERR("Incompatible packet size of Parameter packet!\n");
@@ -1426,12 +1454,29 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 
 	set_bit(MD_DIRTY,&mdev->flags); // we are changing state!
 
-	if( mdev->lo_usize != be64_to_cpu(p->u_size) ) {
-		mdev->lo_usize = be64_to_cpu(p->u_size);
+	p_usize=be64_to_cpu(p->u_size);
+	/*
+	 * you may get a flip-flop connection established/connection loss, in
+	 * case both really have different usize uppon first connect!
+	 * try to solve it thus:
+	 ***/
+#define min_not_zero(l, r) (l == 0) ? r : ((r == 0) ? l : min(l, r))
+	if (mdev->cstate == WFReportParams) {
+		/* this is first connect, or an otherwise expected param
+		 * exchange.  choose the minimum */
+		p_usize = min_not_zero(mdev->lo_usize, p_usize);
+	} else {
+		/* this was an "unexpected" param packet,
+		 * just do what the peer suggests */
+	}
+#undef min_not_zero
+	if( mdev->lo_usize != p_usize ) {
+		mdev->lo_usize = p_usize;
 		INFO("Peer sets u_size to %lu KB\n",
 		     (unsigned long)mdev->lo_usize);
 	}
 
+
 /*lge:
  * FIXME
  * please get the order of tests (re)settings for consider_sync
@@ -1517,8 +1562,12 @@ STATIC int receive_param(drbd_dev *mdev, Drbd_Header *h)
 	oo_state = mdev->o_state;
 	mdev->o_state = be32_to_cpu(p->state);
 	if(oo_state == Secondary && mdev->o_state == Primary) {
-		drbd_md_inc(mdev,ConnectedCnt);
+		/* Secondary has to adopt primary's gen_cnt. */
+		for(i=HumanCnt;i<GEN_CNT_SIZE;i++) {
+			mdev->gen_cnt[i]=be32_to_cpu(p->gen_cnt[i]);
+		}
 	}
+
 	if (oo_state != mdev->o_state) {
 		INFO( "%s/%s --> %s/%s\n",
 		      nodestate_to_name(mdev->state),
@@ -1652,12 +1701,11 @@ STATIC int receive_skip(drbd_dev *mdev,Drbd_Header *h)
 STATIC int receive_BecomeSyncTarget(drbd_dev *mdev, Drbd_Header *h)
 {
 	ERR_IF(!mdev->bitmap) return FALSE;
-
-	/* THINK
-	 * otherwise this does not make much sense, no?
-	 * and some other assertion maybe about cstate...
-	 */
-	ERR_IF(mdev->state != Secondary || mdev->cstate != Connected)
+	ERR_IF(mdev->state != Secondary)
+		return FALSE;
+	ERR_IF(mdev->cstate != Connected)
+		return FALSE;
+	ERR_IF(test_bit(DISKLESS,&mdev->flags))
 		return FALSE;
 
 	drbd_bm_lock(mdev);
@@ -1665,18 +1713,24 @@ STATIC int receive_BecomeSyncTarget(drbd_dev *mdev, Drbd_Header *h)
 	drbd_bm_write(mdev);
 	drbd_start_resync(mdev,SyncTarget);
 	drbd_bm_unlock(mdev);
-	return TRUE; // cannot fail ?
+	return TRUE;
 }
 
 STATIC int receive_BecomeSyncSource(drbd_dev *mdev, Drbd_Header *h)
 {
-	// FIXME asserts ?
+	ERR_IF(mdev->cstate != Connected)
+		return FALSE;
+	ERR_IF(test_bit(DISKLESS,&mdev->flags))
+		return FALSE;
+	ERR_IF(!drbd_md_test_flag(mdev,MDF_Consistent))
+		return FALSE;
+
 	drbd_bm_lock(mdev);
 	drbd_bm_set_all(mdev);
 	drbd_bm_write(mdev);
 	drbd_start_resync(mdev,SyncSource);
 	drbd_bm_unlock(mdev);
-	return TRUE; // cannot fail ?
+	return TRUE;
 }
 
 STATIC int receive_UnplugRemote(drbd_dev *mdev, Drbd_Header *h)
@@ -1852,7 +1906,9 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	if ( mdev->state == Primary &&
 	    ( test_bit(DISKLESS,&mdev->flags)
 	    || !drbd_md_test_flag(mdev,MDF_Consistent) ) ) {
+		drbd_thread_stop_nowait(&mdev->receiver);
 		drbd_panic("Sorry, I have no access to good data anymore.\n");
+		return;
 	}
 
 	if (get_t_state(&mdev->receiver) == Exiting) {
@@ -1876,6 +1932,9 @@ STATIC void drbd_disconnect(drbd_dev *mdev)
 	}
 	clear_bit(DO_NOT_INC_CONCNT,&mdev->flags);
 
+	/* it may still be set, because some unplug was on the fly */
+	NOT_IN_26(mdev->flags &= ~(1<<UNPLUG_QUEUED);)
+
 	INFO("Connection lost.\n");
 }
 
@@ -1906,6 +1965,13 @@ int drbd_send_handshake(drbd_dev *mdev)
 	return ok;
 }
 
+/*
+ * return values:
+ *   1 yess, we have a valid connection
+ *   0 oops, did not work out, please try again
+ *  -1 peer talks different language,
+ *     no point in trying again, please go standalone.
+ */
 STATIC int drbd_do_handshake(drbd_dev *mdev)
 {
 	// ASSERT current == mdev->receiver ...
@@ -1922,17 +1988,17 @@ STATIC int drbd_do_handshake(drbd_dev *mdev)
 	if (p->head.command == ReportParams) {
 		ERR("expected HandShake packet, received ReportParams...\n");
 		ERR("peer probaly runs some incompatible 0.7 -preX version\n");
-		return 0;
+		return -1;
 	} else if (p->head.command != HandShake) {
 		ERR( "expected HandShake packet, received: %s (0x%04x)\n",
 		     cmdname(p->head.command), p->head.command );
-		return 0;
+		return -1;
 	}
 
 	if (p->head.length != expect) {
 		ERR( "expected HandShake length: %u, received: %u\n",
 		     expect, p->head.length );
-		return 0;
+		return -1;
 	}
 
 	rv = drbd_recv(mdev, &p->head.payload, expect);
@@ -1964,7 +2030,7 @@ STATIC int drbd_do_handshake(drbd_dev *mdev)
 		ERR( "incompatible DRBD dialects: "
 		     "I support %u, peer wants %u\n",
 		     PRO_VERSION, p->protocol_version );
-		return 0;
+		return -1;
 	}
 
 	return 1;
@@ -1974,19 +2040,26 @@ int drbdd_init(struct Drbd_thread *thi)
 {
 	drbd_dev *mdev = thi->mdev;
 	int minor = (int)(mdev-drbd_conf);
+	int h;
 
 	sprintf(current->comm, "drbd%d_receiver", minor);
 
 	/* printk(KERN_INFO DEVICE_NAME ": receiver living/m=%d\n", minor); */
 
 	while (TRUE) {
-		if (!drbd_connect(mdev)) {
-			WARN("Discarding network configuration.\n");
+		h = drbd_connect(mdev);
+		if (h <= 0) {
 			/* FIXME DISKLESS StandAlone
 			 * does not make much sense...
 			 * drbd_disconnect should set cstate properly...
 			 */
 			drbd_disconnect(mdev);
+			if (h == 0) {
+				schedule_timeout(HZ);
+				continue;
+			}
+
+			WARN("Discarding network configuration.\n");
 			set_cstate(mdev,StandAlone);
 			break;
 		}
@@ -2049,7 +2122,7 @@ STATIC int got_BlockAck(drbd_dev *mdev, Drbd_Header* h)
 
 		if( is_syncer_blk(mdev,p->block_id)) {
 			drbd_set_in_sync(mdev,sector,blksize);
-			__set_bit(SYNC_STARTED,&mdev->flags);
+			set_bit(SYNC_STARTED,&mdev->flags);
 		} else {
 			req=(drbd_request_t*)(long)p->block_id;
 
@@ -2103,25 +2176,40 @@ STATIC int got_NegAck(drbd_dev *mdev, Drbd_Header* h)
 
 STATIC int got_NegDReply(drbd_dev *mdev, Drbd_Header* h)
 {
-	drbd_request_t *req;
+	/* drbd_request_t *req;
+	 * unused now */
 	Drbd_BlockAck_Packet *p = (Drbd_BlockAck_Packet*)h;
 
+	if (is_syncer_blk(mdev,p->block_id)) {
+		/* no resync data available. don't panic just yet ... */
+		printk(KERN_EMERG DEVICE_NAME "%d: "
+		       "Got NegDReply for resync request. "
+		       "WE ARE LOST. We lost our up-to-date disk.\n",
+			(int)(mdev-drbd_conf));
+		return FALSE;
+	} /* else { */
+
+#if 0
+	/* hey, we panic anyways. so why bother? */
 	req = (drbd_request_t *)(long)p->block_id;
-	D_ASSERT(req->w.cb == w_is_app_read);
+	if (VALID_POINTER(req)) {
+		D_ASSERT(req->w.cb == w_is_app_read);
 
-	spin_lock(&mdev->pr_lock);
-	list_del(&req->w.list);
-	spin_unlock(&mdev->pr_lock);
+		spin_lock(&mdev->pr_lock);
+		list_del(&req->w.list);
+		spin_unlock(&mdev->pr_lock);
 
-	INVALIDATE_MAGIC(req);
-	mempool_free(req,drbd_request_mempool);
+		INVALIDATE_MAGIC(req);
+		mempool_free(req,drbd_request_mempool);
+	}
+#endif
 
 	drbd_panic("Got NegDReply. WE ARE LOST. We lost our up-to-date disk.\n");
 
 	// THINK do we have other options, but panic?
 	//       what about bio_endio, in case we don't panic ??
 
-	return TRUE;
+	return FALSE;
 }
 
 STATIC int got_NegRSDReply(drbd_dev *mdev, Drbd_Header* h)
